MODULE GKV_math
!
!  Mathematical functions
!
!  S. Maeyama     (March 2017)
!

  use GKV_header

  implicit none

  private

  public   math_j0, math_j0zero, math_j1, math_j2, &
           math_i0, math_g0,     math_random

  integer, parameter :: ifnc = 99 ! unit number preserved 
                                  ! for this module


CONTAINS


  SUBROUTINE math_j0( x, j0 )
!
!     0th-order Bessel function
!
    real(kind=DP), intent(in)  :: x
    real(kind=DP), intent(out) :: j0
!
      j0 = dbesj0(x)

    return

  END SUBROUTINE math_j0


  SUBROUTINE math_j0zero( i, j0zero )
!
!     0th-order Bessel function
!
    integer, intent(in)        :: i
    real(kind=DP), intent(out) :: j0zero

    integer, parameter  :: nzero = 4096
    real(kind=DP), save :: j0zeros(nzero)

    integer, save :: isw = 0

      if( i < 1  .or.  i > nzero ) then
        print *, "# range of J0zero is invalid"
        stop
      end if
     
      if( isw == 0 ) then
        include 'Bessel0_Zeros.f90'
      end if

      j0zero = j0zeros(i)
          
    return


  END SUBROUTINE math_j0zero


  SUBROUTINE math_j1( x, j1 )
!
!     1st-order Bessel function
!
    real(kind=DP), intent(in)  :: x
    real(kind=DP), intent(out) :: j1
!
      j1 = dbesj1(x)

    return

  END SUBROUTINE math_j1


  SUBROUTINE math_j2( x, j2 )
!
!     2nd-order Bessel function
!
    real(kind=DP), intent(in)  :: x
    real(kind=DP), intent(out) :: j2
!
      if (x /= 0._DP ) then 
        j2 = 2._DP*dbesj1(x)/x - dbesj0(x)
      else 
        j2 = 0._DP
      end if

    return

  END SUBROUTINE math_j2


  SUBROUTINE math_i0( x, i0 )
!
!     0th-order modified Bessel function
!
    real(kind=DP), intent(in)  :: x
    real(kind=DP), intent(out) :: i0
!
      if( 0._DP <= x  .and.  x < 150._DP ) then
        i0 = dbesi0(x)
      else 
        print *, "### math_i0:  x is out of range!"
      end if

    return

  END SUBROUTINE math_i0


  SUBROUTINE math_g0( x, g0 )
!
!     The Gamma_0 function in gyrokinetics
!     defined by G0(x) = I0(x)*exp(-x)
!
    real(kind=DP), intent(in)  :: x
    real(kind=DP), intent(out) :: g0

    real(kind=DP) :: i0

    real(kind=DP)                 :: twopi
    real(kind=DP), dimension(0:5) :: c

      twopi  = 2._DP * 3.141592653589793_DP
     
      c(0) = 1._DP
      c(1) = 0.25_DP
      c(2) = 9._DP / 32._DP
      c(3) = 75._DP / 128._DP
      c(4) = 3675._DP / 2048._DP
      c(5) = 59535._DP / 8192._DP

      if( x < 150._DP ) then
        call math_i0( x, i0 )
        g0 = i0 * exp( -x )
      else 
        g0 = ( c(0)                      &
             + c(1) / ( 2._DP * x )      &
             + c(2) / ( 2._DP * x )**2   &
             + c(3) / ( 2._DP * x )**3   &
             + c(4) / ( 2._DP * x )**4   &
             + c(5) / ( 2._DP * x )**5 ) &
             / sqrt( twopi * x )
      end if

    return

  END SUBROUTINE math_g0


  SUBROUTINE math_random( rr )
!
!     Random number in [0,1]
!
!
    real(kind=DP), intent(inout), dimension(:) :: rr
!
      call random_number(rr)
!
    return

  END SUBROUTINE math_random


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!%%%                                                                         %%%
!%%%  Ooura's Bessel Functions Library                                       %%%
!%%%                                                                         %%%
!%%%    http://www.kurims.kyoto-u.ac.jp/~ooura/bessel.html                   %%%
!%%%                                                                         %%%
!%%%    Copyright(C) 1996 Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp).   %%%
!%%%    You may use, copy, modify this code for any purpose and              %%%
!%%%    without fee. You may distribute this ORIGINAL package.               %%%
!%%%                                                                         %%%
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

! Bessel J_0(x) function in double precision
!
      function dbesj0(x)
      implicit integer (i - n)
      implicit real*8 (a - h, o - z)
      dimension a(0 : 7), b(0 : 64), c(0 : 69), d(0 : 51)
      parameter (pi4 = 0.78539816339744830962d0)
      data (a(i), i = 0, 7) /                                     &
     &    -0.0000000000023655394d0, 0.0000000004708898680d0,      &
     &    -0.0000000678167892231d0, 0.0000067816840038636d0,      &
     &    -0.0004340277777716935d0, 0.0156249999999992397d0,      &
     &    -0.2499999999999999638d0, 0.9999999999999999997d0 / 
      data (b(i), i = 0, 12) /                                    &
     &    0.0000000000626681117d0, -0.0000000022270614428d0,      &
     &    0.0000000662981656302d0, -0.0000016268486502196d0,      &
     &    0.0000321978384111685d0, -0.0005005237733315830d0,      &
     &    0.0059060313537449816d0, -0.0505265323740109701d0,      &
     &    0.2936432097610503985d0, -1.0482565081091638637d0,      &
     &    1.9181123286040428113d0, -1.1319199475221700100d0,      &
     &    -0.1965480952704682000d0 / 
      data (b(i), i = 13, 25) /                                   &
     &    0.0000000000457457332d0, -0.0000000015814772025d0,      &
     &    0.0000000455487446311d0, -0.0000010735201286233d0,      &
     &    0.0000202015179970014d0, -0.0002942392368203808d0,      &
     &    0.0031801987726150648d0, -0.0239875209742846362d0,      &
     &    0.1141447698973777641d0, -0.2766726722823530233d0,      &
     &    0.1088620480970941648d0, 0.5136514645381999197d0,       &
     &    -0.2100594022073706033d0 / 
      data (b(i), i = 26, 38) /                                   &
     &    0.0000000000331366618d0, -0.0000000011119090229d0,      &
     &    0.0000000308823040363d0, -0.0000006956602653104d0,      &
     &    0.0000123499947481762d0, -0.0001662951945396180d0,      &
     &    0.0016048663165678412d0, -0.0100785479932760966d0,      &
     &    0.0328996815223415274d0, -0.0056168761733860688d0,      &
     &    -0.2341096400274429386d0, 0.2551729256776404262d0,      &
     &    0.2288438186148935667d0 / 
      data (b(i), i = 39, 51) /                                   &
     &    0.0000000000238007203d0, -0.0000000007731046439d0,      &
     &    0.0000000206237001152d0, -0.0000004412291442285d0,      &
     &    0.0000073107766249655d0, -0.0000891749801028666d0,      &
     &    0.0007341654513841350d0, -0.0033303085445352071d0,      &
     &    0.0015425853045205717d0, 0.0521100583113136379d0,       &
     &    -0.1334447768979217815d0, -0.1401330292364750968d0,     &
     &    0.2685616168804818919d0 / 
      data (b(i), i = 52, 64) /                                   &
     &    0.0000000000169355950d0, -0.0000000005308092192d0,      &
     &    0.0000000135323005576d0, -0.0000002726650587978d0,      &
     &    0.0000041513240141760d0, -0.0000443353052220157d0,      &
     &    0.0002815740758993879d0, -0.0004393235121629007d0,      &
     &    -0.0067573531105799347d0, 0.0369141914660130814d0,      &
     &    0.0081673361942996237d0, -0.2573381285898881860d0,      &
     &    0.0459580257102978932d0 / 
      data (c(i), i = 0, 13) /                                    &
     &    -0.00000000003009451757d0, -0.00000000014958003844d0,   & 
     &    0.00000000506854544776d0, 0.00000001863564222012d0,     &
     &    -0.00000060304249068078d0, -0.00000147686259937403d0,   & 
     &    0.00004714331342682714d0, 0.00006286305481740818d0,     &
     &    -0.00214137170594124344d0, -0.00089157336676889788d0,   &
     &    0.04508258728666024989d0, -0.00490362805828762224d0,    &
     &    -0.27312196367405374426d0, 0.04193925184293450356d0 /   
      data (c(i), i = 14, 27) /                                   &
     &    -0.00000000000712453560d0, -0.00000000041170814825d0,   &
     &    0.00000000138012624364d0, 0.00000005704447670683d0,     &
     &    -0.00000019026363528842d0, -0.00000533925032409729d0,   &
     &    0.00001736064885538091d0, 0.00030692619152608375d0,     &
     &    -0.00092598938200644367d0, -0.00917934265960017663d0,   &
     &    0.02287952522866389076d0, 0.10545197546252853195d0,     &
     &    -0.16126443075752985095d0, -0.19392874768742235538d0 /  
      data (c(i), i = 28, 41) /                                   &
     &    0.00000000002128344556d0, -0.00000000031053910272d0,    &
     &    -0.00000000334979293158d0, 0.00000004507232895050d0,    &
     &    0.00000036437959146427d0, -0.00000446421436266678d0,    &
     &    -0.00002523429344576552d0, 0.00027519882931758163d0,    &
     &    0.00097185076358599358d0, -0.00898326746345390692d0,    &
     &    -0.01665959196063987584d0, 0.11456933464891967814d0,    &
     &    0.07885001422733148815d0, -0.23664819446234712621d0 /   
      data (c(i), i = 42, 55) /                                   &
     &    0.00000000003035295055d0, 0.00000000005486066835d0,     &
     &    -0.00000000501026824811d0, -0.00000000501246847860d0,   &
     &    0.00000058012340163034d0, 0.00000016788922416169d0,     &
     &    -0.00004373270270147275d0, 0.00001183898532719802d0,    &
     &    0.00189863342862291449d0, -0.00113759249561636130d0,    &
     &    -0.03846797195329871681d0, 0.02389746880951420335d0,    &
     &    0.22837862066532347461d0, -0.06765394811166522844d0 /   
      data (c(i), i = 56, 69) /                                   &
     &    0.00000000001279875977d0, 0.00000000035925958103d0,     &
     &    -0.00000000228037105967d0, -0.00000004852770517176d0,   &
     &    0.00000028696428000189d0, 0.00000440131125178642d0,     &
     &    -0.00002366617753349105d0, -0.00024412456252884129d0,   &
     &    0.00113028178539430542d0, 0.00708470513919789080d0,     &
     &    -0.02526914792327618386d0, -0.08006137953480093426d0,   &
     &    0.16548380461475971846d0, 0.14688405470042110229d0 /   
      data (d(i), i = 0, 12) /                                    &
     &    1.059601355592185731d-14, -2.71150591218550377d-13,     &
     &    8.6514809056201638d-12, -4.6264028554286627d-10,        &
     &    5.0815403835647104d-8, -1.76722552048141208d-5,         &
     &    0.16286750396763997378d0, 2.949651820598278873d-13,     &
     &    -8.818215611676125741d-12, 3.571119876162253451d-10,    &
     &    -2.631924120993717060d-8, 4.709502795656698909d-6,      &
     &    -5.208333333333283282d-3 / 
      data (d(i), i = 13, 25) /                                   &
     &    7.18344107717531977d-15, -2.51623725588410308d-13,      &
     &    8.6017784918920604d-12, -4.6256876614290359d-10,        &
     &    5.0815343220437937d-8, -1.76722551764941970d-5,         &
     &    0.16286750396763433767d0, 2.2327570859680094777d-13,    &
     &    -8.464594853517051292d-12, 3.563766464349055183d-10,    &
     &    -2.631843986737892965d-8, 4.709502342288659410d-6,      &
     &    -5.2083333332278466225d-3 / 
      data (d(i), i = 26, 38) /                                   &
     &    5.15413392842889366d-15, -2.27740238380640162d-13,      &
     &    8.4827767197609014d-12, -4.6224753682737618d-10,        &
     &    5.0814848128929134d-8, -1.76722547638767480d-5,         &
     &    0.16286750396748926663d0, 1.7316195320192170887d-13,    &
     &    -7.971122772293919646d-12, 3.544039469911895749d-10,    &
     &    -2.631443902081701081d-8, 4.709498228695400603d-6,      &
     &    -5.2083333315143653610d-3 / 
      data (d(i), i = 39, 51) /                                   &
     &    3.84653681453798517d-15, -2.04464520778789011d-13,      &
     &    8.3089298605177838d-12, -4.6155016158412096d-10,        &
     &    5.0813263696466650d-8, -1.76722528311426167d-5,         &
     &    0.16286750396650065930d0, 1.3797879972460878797d-13,    &
     &    -7.448089381011684812d-12, 3.512733797106959780d-10,    &
     &    -2.630500895563592722d-8, 4.709483934775839193d-6,      &
     &    -5.2083333227940760113d-3 / 
      w = abs(x)
      if (w .lt. 1) then
          t = w * w
          y = ((((((a(0) * t + a(1)) * t +                        &
     &        a(2)) * t + a(3)) * t + a(4)) * t +                 &
     &        a(5)) * t + a(6)) * t + a(7)
      else if (w .lt. 8.5d0) then
          t = w * w * 0.0625d0
          k = int(t)
          t = t - (k + 0.5d0)
          k = k * 13
          y = (((((((((((b(k) * t + b(k + 1)) * t +               &
     &        b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t +     &
     &        b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t +     &
     &        b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t +    &
     &        b(k + 11)) * t + b(k + 12)
      else if (w .lt. 12.5d0) then
          k = int(w)
          t = w - (k + 0.5d0)
          k = 14 * (k - 8)
          y = ((((((((((((c(k) * t + c(k + 1)) * t +              &
     &        c(k + 2)) * t + c(k + 3)) * t + c(k + 4)) * t +     &
     &        c(k + 5)) * t + c(k + 6)) * t + c(k + 7)) * t +     &
     &        c(k + 8)) * t + c(k + 9)) * t + c(k + 10)) * t +    &
     &        c(k + 11)) * t + c(k + 12)) * t + c(k + 13)
      else
          v = 24 / w
          t = v * v
          k = 13 * (int(t))
          y = ((((((d(k) * t + d(k + 1)) * t +                    &
     &        d(k + 2)) * t + d(k + 3)) * t + d(k + 4)) * t +     &
     &        d(k + 5)) * t + d(k + 6)) * sqrt(v)
          theta = (((((d(k + 7) * t + d(k + 8)) * t +             &
     &        d(k + 9)) * t + d(k + 10)) * t + d(k + 11)) * t +   &
     &        d(k + 12)) * v - pi4
          y = y * cos(w + theta)
      end if
      dbesj0 = y
      end function dbesj0
!
! Bessel J_1(x) function in double precision
!
      function dbesj1(x)
      implicit integer (i - n)
      implicit real*8 (a - h, o - z)
      dimension a(0 : 7), b(0 : 64), c(0 : 69), d(0 : 51)
      parameter (pi4 = 0.78539816339744830962d0)
      data (a(i), i = 0, 7) /                                     &
     &    -0.00000000000014810349d0, 0.00000000003363594618d0,    &
     &    -0.00000000565140051697d0, 0.00000067816840144764d0,    &
     &    -0.00005425347222188379d0, 0.00260416666666662438d0,    &
     &    -0.06249999999999999799d0, 0.49999999999999999998d0 /   
      data (b(i), i = 0, 12) /                                    &
     &    0.00000000000243721316d0, -0.00000000009400554763d0,    &
     &    0.00000000306053389980d0, -0.00000008287270492518d0,    &
     &    0.00000183020515991344d0, -0.00003219783841164382d0,    &
     &    0.00043795830161515318d0, -0.00442952351530868999d0,    &
     &    0.03157908273375945955d0, -0.14682160488052520107d0,    &
     &    0.39309619054093640008d0, -0.47952808215101070280d0,    &
     &    0.14148999344027125140d0 /                              
      data (b(i), i = 13, 25) /                                   &
     &    0.00000000000182119257d0, -0.00000000006862117678d0,    &
     &    0.00000000217327908360d0, -0.00000005693592917820d0,    &
     &    0.00000120771046483277d0, -0.00002020151799736374d0,    &
     &    0.00025745933218048448d0, -0.00238514907946126334d0,    &
     &    0.01499220060892984289d0, -0.05707238494868888345d0,    &
     &    0.10375225210588234727d0, -0.02721551202427354117d0,    &
     &    -0.06420643306727498985d0 /                             
      data (b(i), i = 26, 38) /                                   &
     &    0.000000000001352611196d0, -0.000000000049706947875d0,  &
     &    0.000000001527944986332d0, -0.000000038602878823401d0,  &
     &    0.000000782618036237845d0, -0.000012349994748451100d0,  &
     &    0.000145508295194426686d0, -0.001203649737425854162d0,  &
     &    0.006299092495799005109d0, -0.016449840761170764763d0,  &
     &    0.002106328565019748701d0, 0.058527410006860734650d0,   &
     &    -0.031896615709705053191d0 /                            
      data (b(i), i = 39, 51) /                                   &
     &    0.000000000000997982124d0, -0.000000000035702556073d0,  &
     &    0.000000001062332772617d0, -0.000000025779624221725d0,  &
     &    0.000000496382962683556d0, -0.000007310776625173004d0,  &
     &    0.000078028107569541842d0, -0.000550624088538081113d0,  &
     &    0.002081442840335570371d0, -0.000771292652260286633d0,  &
     &    -0.019541271866742634199d0, 0.033361194224480445382d0,  &
     &    0.017516628654559387164d0 /                             
      data (b(i), i = 52, 64) /                                   &
     &    0.000000000000731050661d0, -0.000000000025404499912d0,  &
     &    0.000000000729360079088d0, -0.000000016915375004937d0,  &
     &    0.000000306748319652546d0, -0.000004151324014331739d0,  &
     &    0.000038793392054271497d0, -0.000211180556924525773d0,  &
     &    0.000274577195102593786d0, 0.003378676555289966782d0,   &
     &    -0.013842821799754920148d0, -0.002041834048574905921d0, &
     &    0.032167266073736023299d0 /                             
      data (c(i), i = 0, 13) /                                    &
     &    -0.00000000001185964494d0, 0.00000000039110295657d0,    &
     &    0.00000000180385519493d0, -0.00000005575391345723d0,    &
     &    -0.00000018635897017174d0, 0.00000542738239401869d0,    &
     &    0.00001181490114244279d0, -0.00033000319398521070d0,    &
     &    -0.00037717832892725053d0, 0.01070685852970608288d0,    &
     &    0.00356629346707622489d0, -0.13524776185998074716d0,    &
     &    0.00980725611657523952d0, 0.27312196367405374425d0 /    
      data (c(i), i = 14, 27) /                                   &
     &    -0.00000000003029591097d0, 0.00000000009259293559d0,    &
     &    0.00000000496321971223d0, -0.00000001518137078639d0,    &
     &    -0.00000057045127595547d0, 0.00000171237271302072d0,    &
     &    0.00004271400348035384d0, -0.00012152454198713258d0,    &
     &    -0.00184155714921474963d0, 0.00462994691003219055d0,    &
     &    0.03671737063840232452d0, -0.06863857568599167175d0,    &
     &    -0.21090395092505707655d0, 0.16126443075752985095d0 /   
      data (c(i), i = 28, 41) /                                   &
     &    -0.00000000002197602080d0, -0.00000000027659100729d0,   &
     &    0.00000000374295124827d0, 0.00000003684765777023d0,     &
     &    -0.00000045072801091574d0, -0.00000327941630669276d0,   &
     &    0.00003571371554516300d0, 0.00017664005411843533d0,     &
     &    -0.00165119297594774104d0, -0.00485925381792986774d0,   &
     &    0.03593306985381680131d0, 0.04997877588191962563d0,     &
     &    -0.22913866929783936544d0, -0.07885001422733148814d0 /  
      data (c(i), i = 42, 55) /                                   &
     &    0.00000000000516292316d0, -0.00000000039445956763d0,    &
     &    -0.00000000066220021263d0, 0.00000005511286218639d0,    &
     &    0.00000005012579400780d0, -0.00000522111059203425d0,    &
     &    -0.00000134311394455105d0, 0.00030612891890766805d0,    &
     &    -0.00007103391195326182d0, -0.00949316714311443491d0,   &
     &    0.00455036998246516948d0, 0.11540391585989614784d0,     &
     &    -0.04779493761902840455d0, -0.22837862066532347460d0 /  
      data (c(i), i = 56, 69) /                                   &
     &    0.00000000002697817493d0, -0.00000000016633326949d0,    &
     &    -0.00000000433134860350d0, 0.00000002508404686362d0,    &
     &    0.00000048528284780984d0, -0.00000258267851112118d0,    &
     &    -0.00003521049080466759d0, 0.00016566324273339952d0,    &
     &    0.00146474737522491617d0, -0.00565140892697147306d0,    &
     &    -0.02833882055679300400d0, 0.07580744376982855057d0,    &
     &    0.16012275906960187978d0, -0.16548380461475971845d0 /   
      data (d(i), i = 0, 12) /                                    &
     &    -1.272346002224188092d-14, 3.370464692346669075d-13,    &
     &    -1.144940314335484869d-11, 6.863141561083429745d-10,    &
     &    -9.491933932960924159d-8, 5.301676561445687562d-5,      &
     &    0.1628675039676399740d0, -3.652982212914147794d-13,     &
     &    1.151126750560028914d-11, -5.165585095674343486d-10,    &
     &    4.657991250060549892d-8, -1.186794704692706504d-5,      &
     &    1.562499999999994026d-2 /                               
      data (d(i), i = 13, 25) /                                   &
     &    -8.713069680903981555d-15, 3.140780373478474935d-13,    &
     &    -1.139089186076256597d-11, 6.862299023338785566d-10,    &
     &    -9.491926788274594674d-8, 5.301676558106268323d-5,      &
     &    0.1628675039676466220d0, -2.792555727162752006d-13,     &
     &    1.108650207651756807d-11, -5.156745588549830981d-10,    &
     &    4.657894859077370979d-8, -1.186794650130550256d-5,      &
     &    1.562499999987299901d-2 /                               
      data (d(i), i = 26, 38) /                                   &
     &    -6.304859171204770696d-15, 2.857249044208791652d-13,    &
     &    -1.124956921556753188d-11, 6.858482894906716661d-10,    &
     &    -9.491867953516898460d-8, 5.301676509057781574d-5,      &
     &    0.1628675039678191167d0, -2.185193490132496053d-13,     &
     &    1.048820673697426074d-11, -5.132819367467680132d-10,    &
     &    4.657409437372994220d-8, -1.186794150862988921d-5,      &
     &    1.562499999779270706d-2 /                               
      data (d(i), i = 39, 51) /                                   &
     &    -4.740417209792009850d-15, 2.578715253644144182d-13,    &
     &    -1.104148898414138857d-11, 6.850134201626289183d-10,    &
     &    -9.491678234174919640d-8, 5.301676277588728159d-5,      &
     &    0.1628675039690033136d0, -1.755122057493842290d-13,     &
     &    9.848723331445182397d-12, -5.094535425482245697d-10,    &
     &    4.656255982268609304d-8, -1.186792402114394891d-5,      &
     &    1.562499998712198636d-2 / 
      w = abs(x)
      if (w .lt. 1) then
          t = w * w
          y = (((((((a(0) * t + a(1)) * t +                       &
     &        a(2)) * t + a(3)) * t + a(4)) * t +                 &
     &        a(5)) * t + a(6)) * t + a(7)) * w
      else if (w .lt. 8.5d0) then
          t = w * w * 0.0625d0
          k = int(t)
          t = t - (k + 0.5d0)
          k = k * 13
          y = ((((((((((((b(k) * t + b(k + 1)) * t +              &
     &        b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t +     &
     &        b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t +     &
     &        b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t +    &
     &        b(k + 11)) * t + b(k + 12)) * w
      else if (w .lt. 12.5d0) then
          k = int(w)
          t = w - (k + 0.5d0)
          k = 14 * (k - 8)
          y = ((((((((((((c(k) * t + c(k + 1)) * t +              &
     &        c(k + 2)) * t + c(k + 3)) * t + c(k + 4)) * t +     &
     &        c(k + 5)) * t + c(k + 6)) * t + c(k + 7)) * t +     &
     &        c(k + 8)) * t + c(k + 9)) * t + c(k + 10)) * t +    &
     &        c(k + 11)) * t + c(k + 12)) * t + c(k + 13)
      else
          v = 24 / w
          t = v * v
          k = 13 * (int(t))
          y = ((((((d(k) * t + d(k + 1)) * t +                    &
     &        d(k + 2)) * t + d(k + 3)) * t + d(k + 4)) * t +     &
     &        d(k + 5)) * t + d(k + 6)) * sqrt(v)                 
          theta = (((((d(k + 7) * t + d(k + 8)) * t +             &
     &        d(k + 9)) * t + d(k + 10)) * t + d(k + 11)) * t +   &
     &        d(k + 12)) * v - pi4
          y = y * sin(w + theta)
      end if
      if (x .lt. 0) y = -y
      dbesj1 = y
      end function dbesj1
!
! Bessel I_0(x) function in double precision
!
      function dbesi0(x)
      implicit integer (i - n)
      implicit real*8 (a - h, o - z)
      dimension a(0 : 64), b(0 : 69), c(0 : 44)
      data (a(i), i = 0, 12) /                                    &
     &    8.5246820682016865877d-11, 2.5966600546497407288d-9,    &
     &    7.9689994568640180274d-8, 1.9906710409667748239d-6,     &
     &    4.0312469446528002532d-5, 6.4499871606224265421d-4,     &
     &    7.9012345761930579108d-3, 7.1111111109207045212d-2,     &
     &    4.4444444444472490900d-1, 1.7777777777777532045d0,      &
     &    4.0000000000000011182d0, 3.9999999999999999800d0,       &
     &    1.0000000000000000001d0 /                                
      data (a(i), i = 13, 25) /                                   &
     &    1.1520919130377195927d-10, 2.2287613013610985225d-9,    &
     &    8.1903951930694585113d-8, 1.9821560631611544984d-6,     &
     &    4.0335461940910133184d-5, 6.4495330974432203401d-4,     &
     &    7.9013012611467520626d-3, 7.1111038160875566622d-2,     &
     &    4.4444450319062699316d-1, 1.7777777439146450067d0,      &
     &    4.0000000132337935071d0, 3.9999999968569015366d0,       &
     &    1.0000000003426703174d0 /                                
      data (a(i), i = 26, 38) /                                   &
     &    1.5476870780515238488d-10, 1.2685004214732975355d-9,    &
     &    9.2776861851114223267d-8, 1.9063070109379044378d-6,     &
     &    4.0698004389917945832d-5, 6.4370447244298070713d-4,     &
     &    7.9044749458444976958d-3, 7.1105052411749363882d-2,     &
     &    4.4445280640924755082d-1, 1.7777694934432109713d0,      &
     &    4.0000055808824003386d0, 3.9999977081165740932d0,       &
     &    1.0000004333949319118d0 /                                
      data (a(i), i = 39, 51) /                                   &
     &    2.0675200625006793075d-10, -6.1689554705125681442d-10,  &
     &    1.2436765915401571654d-7, 1.5830429403520613423d-6,     &
     &    4.2947227560776583326d-5, 6.3249861665073441312d-4,     &
     &    7.9454472840953930811d-3, 7.0994327785661860575d-2,     &
     &    4.4467219586283000332d-1, 1.7774588182255374745d0,      &
     &    4.0003038986252717972d0, 3.9998233869142057195d0,       &
     &    1.0000472932961288324d0 /                                
      data (a(i), i = 52, 64) /                                   &
     &    2.7475684794982708655d-10, -3.8991472076521332023d-9,   &
     &    1.9730170483976049388d-7, 5.9651531561967674521d-7,     &
     &    5.1992971474748995357d-5, 5.7327338675433770752d-4,     &
     &    8.2293143836530412024d-3, 6.9990934858728039037d-2,     &
     &    4.4726764292723985087d-1, 1.7726685170014087784d0,      &
     &    4.0062907863712704432d0, 3.9952750700487845355d0,       &
     &    1.0016354346654179322d0 /                                
      data (b(i), i = 0, 13) /                                    &
     &    6.7852367144945531383d-8, 4.6266061382821826854d-7,     &
     &    6.9703135812354071774d-6, 7.6637663462953234134d-5,     &
     &    7.9113515222612691636d-4, 7.3401204731103808981d-3,     &
     &    6.0677114958668837046d-2, 4.3994941411651569622d-1,     &
     &    2.7420017097661750609d0, 14.289661921740860534d0,       &
     &    59.820609640320710779d0, 188.78998681199150629d0,       &
     &    399.87313678256011180d0, 427.56411572180478514d0 /       
      data (b(i), i = 14, 27) /                                   &
     &    1.8042097874891098754d-7, 1.2277164312044637357d-6,     &
     &    1.8484393221474274861d-5, 2.0293995900091309208d-4,     &
     &    2.0918539850246207459d-3, 1.9375315654033949297d-2,     &
     &    1.5985869016767185908d-1, 1.1565260527420641724d0,      &
     &    7.1896341224206072113d0, 37.354773811947484532d0,       &
     &    155.80993164266268457d0, 489.52113711585409180d0,       &
     &    1030.9147225169564806d0, 1093.5883545113746958d0 /       
      data (b(i), i = 28, 41) /                                   &
     &    4.8017305613187493564d-7, 3.2613178439123800740d-6,     &
     &    4.9073137508166159639d-5, 5.3806506676487583755d-4,     &
     &    5.5387918291051866561d-3, 5.1223717488786549025d-2,     &
     &    4.2190298621367914765d-1, 3.0463625987357355872d0,      &
     &    18.895299447327733204d0, 97.915189029455461554d0,       &
     &    407.13940115493494659d0, 1274.3088990480582632d0,       &
     &    2670.9883037012547506d0, 2815.7166284662544712d0 /       
      data (b(i), i = 42, 55) /                                   &
     &    1.2789926338424623394d-6, 8.6718263067604918916d-6,     &
     &    1.3041508821299929489d-4, 1.4282247373727478920d-3,     &
     &    1.4684070635768789378d-2, 1.3561403190404185755d-1,     &
     &    1.1152592585977393953d0, 8.0387088559465389038d0,       &
     &    49.761318895895479206d0, 257.26842323135291380d0,       &
     &    1066.8543146269566231d0, 3328.3874581009636362d0,       &
     &    6948.8586598121634874d0, 7288.4893398212481055d0 /       
      data (b(i), i = 56, 69) /                                   &
     &    3.4093503681970328930d-6, 2.3079025203103376076d-5,     &
     &    3.4691373283901830239d-4, 3.7949949772229085450d-3,     &
     &    3.8974209677945602145d-2, 3.5949483804148783710d-1,     &
     &    2.9522878893539528226d0, 21.246564609514287056d0,       &
     &    131.28727387146173141d0, 677.38107093296675421d0,       &
     &    2802.3724744545046518d0, 8718.5731420798254081d0,       &
     &    18141.348781638832286d0, 18948.925349296308859d0 /       
      data (c(i), i = 0, 8) /                                     &
     &    2.5568678676452702768d-15, 3.0393953792305924324d-14,   &
     &    6.3343751991094840009d-13, 1.5041298011833009649d-11,   &
     &    4.4569436918556541414d-10, 1.7463930514271679510d-8,    &
     &    1.0059224011079852317d-6, 1.0729838945088577089d-4,     &
     &    5.1503226936425277380d-2 /                               
      data (c(i), i = 9, 17) /                                    &
     &    5.2527963991711562216d-15, 7.2021184814210056410d-15,   &
     &    7.2561421229904797156d-13, 1.4823121466731042510d-11,   &
     &    4.4602670450376245434d-10, 1.7463600061788679671d-8,    &
     &    1.0059226091322347560d-6, 1.0729838937545111487d-4,     &
     &    5.1503226936437300716d-2 /                               
      data (c(i), i = 18, 26) /                                   &
     &    1.3365917359358069908d-14, -1.2932643065888544835d-13,  &
     &    1.7450199447905602915d-12, 1.0419051209056979788d-11,   &
     &    4.5804788198059832600d-10, 1.7442405450073548966d-8,    &
     &    1.0059461453281292278d-6, 1.0729837434500161228d-4,     &
     &    5.1503226940658446941d-2 /                               
      data (c(i), i = 27, 35) /                                   &
     &    5.3771611477352308649d-14, -1.1396193006413731702d-12,  &
     &    1.2858641335221653409d-11, -5.9802086004570057703d-11,  &
     &    7.3666894305929510222d-10, 1.6731837150730356448d-8,    &
     &    1.0070831435812128922d-6, 1.0729733111203704813d-4,     &
     &    5.1503227360726294675d-2 /                               
      data (c(i), i = 36, 44) /                                   &
     &    3.7819492084858931093d-14, -4.8600496888588034879d-13,  &
     &    1.6898350504817224909d-12, 4.5884624327524255865d-11,   &
     &    1.2521615963377513729d-10, 1.8959658437754727957d-8,    &
     &    1.0020716710561353622d-6, 1.0730371198569275590d-4,     &
     &    5.1503223833002307750d-2 / 
      w = abs(x)
      if (w .lt. 8.5d0) then
          t = w * w * 0.0625d0
          k = 13 * (int(t))
          y = (((((((((((a(k) * t + a(k + 1)) * t +               &
     &        a(k + 2)) * t + a(k + 3)) * t + a(k + 4)) * t +     &
     &        a(k + 5)) * t + a(k + 6)) * t + a(k + 7)) * t +     &
     &        a(k + 8)) * t + a(k + 9)) * t + a(k + 10)) * t +    &
     &        a(k + 11)) * t + a(k + 12)
      else if (w .lt. 12.5d0) then
          k = int(w)
          t = w - k
          k = 14 * (k - 8)
          y = ((((((((((((b(k) * t + b(k + 1)) * t +              &
     &        b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t +     &
     &        b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t +     &
     &        b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t +    &
     &        b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
      else
          t = 60 / w
          k = 9 * (int(t))
          y = ((((((((c(k) * t + c(k + 1)) * t +                  &
     &        c(k + 2)) * t + c(k + 3)) * t + c(k + 4)) * t +     &
     &        c(k + 5)) * t + c(k + 6)) * t + c(k + 7)) * t +     &
     &        c(k + 8)) * sqrt(t) * exp(w)
      end if
      dbesi0 = y
      end function dbesi0
!
! Bessel I_1(x) function in double precision
!
      function dbesi1(x)
      implicit integer (i - n)
      implicit real*8 (a - h, o - z)
      dimension a(0 : 59), b(0 : 69), c(0 : 44)
      data (a(i), i = 0, 11) /                                    &
     &    1.2787464404046789181d-10, 3.5705860060088241077d-9,    &
     &    9.9611537619347335040d-8, 2.2395070088633043177d-6,     &
     &    4.0312466928887462346d-5, 5.6437387840203722356d-4,     &
     &    5.9259259312934746096d-3, 4.4444444443499008870d-2,     &
     &    2.2222222222232042719d-1, 6.6666666666666139867d-1,     &
     &    1.0000000000000001106d0, 4.9999999999999999962d-1 /      
      data (a(i), i = 12, 23) /                                   &
     &    1.7281952384448634449d-10, 3.0647204559976390130d-9,    &
     &    1.0237662138842827028d-7, 2.2299494417341498163d-6,     &
     &    4.0335364374929326943d-5, 5.6433440269141349899d-4,     &
     &    5.9259754885893798654d-3, 4.4444399410880397870d-2,     &
     &    2.2222225112835026730d-1, 6.6666665422146063244d-1,     &
     &    1.0000000032274936821d0, 4.9999999961866867205d-1 /      
      data (a(i), i = 24, 35) /                                   &
     &    2.3216048939948030996d-10, 1.7443372702334489579d-9,    &
     &    1.1596478963485415499d-7, 2.1446755518623035147d-6,     &
     &    4.0697440347437076195d-5, 5.6324394900433192204d-4,     &
     &    5.9283484996093060678d-3, 4.4440673899150997921d-2,     &
     &    2.2222638016852657860d-1, 6.6666358151576732094d-1,     &
     &    1.0000013834029985337d0, 4.9999971643129650249d-1 /      
      data (a(i), i = 36, 47) /                                   &
     &    3.1013758938255172562d-10, -8.4813676145611694984d-10,  &
     &    1.5544980187411802596d-7, 1.7811109378708045726d-6,     &
     &    4.2945322199060856985d-5, 5.5344850176852353639d-4,     &
     &    5.9590327716950614802d-3, 4.4371611097707060659d-2,     &
     &    2.2233578241986401111d-1, 6.6654747300463315310d-1,     &
     &    1.0000756505206705927d0, 4.9997803664415994554d-1 /      
      data (a(i), i = 48, 59) /                                   &
     &    4.1214758313965020365d-10, -5.3613317735347429440d-9,   &
     &    2.4661360807517345161d-7, 6.7144593918926723203d-7,     &
     &    5.1988027944945587571d-5, 5.0165568586065803067d-4,     &
     &    6.1717530047005289953d-3, 4.3745229577317251404d-2,     &
     &    2.2363147971477747996d-1, 6.6475469131117660240d-1,     &
     &    1.0015686689447547657d0, 4.9941120439785391891d-1 /      
      data (b(i), i = 0, 13) /                                    &
     &    6.6324787943143095845d-8, 4.5125928898466638619d-7,     &
     &    6.7937793134877246623d-6, 7.4580507871505926302d-5,     &
     &    7.6866382927334005919d-4, 7.1185174803491859307d-3,     &
     &    5.8721838073486424416d-2, 4.2473949281714196041d-1,     &
     &    2.6396965606282079123d0, 13.710008536637016903d0,       &
     &    57.158647688180932003d0, 179.46182892089389037d0,       &
     &    377.57997362398478619d0, 399.87313678256009819d0 /       
      data (b(i), i = 14, 27) /                                   &
     &    1.7652713206027939711d-7, 1.1988179244834708057d-6,     &
     &    1.8037851545747139231d-5, 1.9775785516370314656d-4,     &
     &    2.0354870702829387283d-3, 1.8822164191032253600d-2,     &
     &    1.5500485219010424263d-1, 1.1190100010560573210d0,      &
     &    6.9391565185406617552d0, 35.948170579648649345d0,       &
     &    149.41909525103032616d0, 467.42979492780642582d0,       &
     &    979.04227423171290408d0, 1030.9147225169564443d0 /       
      data (b(i), i = 28, 41) /                                   &
     &    4.7022299276154507603d-7, 3.1878571710170115972d-6,     &
     &    4.7940153875711448496d-5, 5.2496623508411440227d-4,     &
     &    5.3968661134780824779d-3, 4.9837081920693776234d-2,     &
     &    4.0979593830387765545d-1, 2.9533186922862948404d0,      &
     &    18.278176130722516369d0, 94.476497150189121070d0,       &
     &    391.66075612645333624d0, 1221.4182034643210345d0,       &
     &    2548.6177980961291004d0, 2670.9883037012546541d0 /       
      data (b(i), i = 42, 55) /                                   &
     &    1.2535083724002034147d-6, 8.4845871420655708250d-6,     &
     &    1.2753227372734042108d-4, 1.3950105363562648921d-3,     &
     &    1.4325473993765291906d-2, 1.3212452778932829125d-1,     &
     &    1.0849287786885151432d0, 7.8068089156260172673d0,       &
     &    48.232254570679165833d0, 248.80659424902394371d0,       &
     &    1029.0736929484210803d0, 3200.5629438795801652d0,       &
     &    6656.7749162019607914d0, 6948.8586598121632302d0 /       
      data (b(i), i = 56, 69) /                                   &
     &    3.3439394490599745013d-6, 2.2600596902211837757d-5,     &
     &    3.3955927589987356838d-4, 3.7105306061050972474d-3,     &
     &    3.8065263634919156421d-2, 3.5068223415665236079d-1,     &
     &    2.8760027832105027316d0, 20.665999500843274339d0,       &
     &    127.47939148516390205d0, 656.43636874254000885d0,       &
     &    2709.5242837932479920d0, 8407.1174233600734871d0,       &
     &    17437.146284159740233d0, 18141.348781638831600d0 /       
      data (c(i), i = 0, 8) /                                     &
     &    -2.8849790431465382128d-15, -3.5125350943844774657d-14, &
     &    -7.4850867013707419750d-13, -1.8383904048277485153d-11, &
     &    -5.7303556446977223342d-10, -2.4449502737311496525d-8,  &
     &    -1.6765373351766929724d-6, -3.2189516835265773471d-4,   &
     &    5.1503226936425277377d-2 /                               
      data (c(i), i = 9, 17) /                                    &
     &    -5.8674306822281631119d-15, -9.4884898451194085565d-15, &
     &    -8.5033865136600364340d-13, -1.8142997866945285736d-11, &
     &    -5.7340238386338193949d-10, -2.4449138101742183665d-8,  &
     &    -1.6765375646678855842d-6, -3.2189516826945356325d-4,   &
     &    5.1503226936412017608d-2 /                               
      data (c(i), i = 18, 26) /                                   &
     &    -1.4723362506764340882d-14, 1.3945147385179042899d-13,  &
     &    -1.9618041857586930923d-12, -1.3343606394065121821d-11, &
     &    -5.8649674606973244159d-10, -2.4426060539669553778d-8,  &
     &    -1.6765631828366988006d-6, -3.2189515191449587253d-4,   &
     &    5.1503226931820146445d-2 /                               
      data (c(i), i = 27, 35) /                                   &
     &    -5.8203519372580372987d-14, 1.2266326995309845825d-12,  &
     &    -1.3921625844526453237d-11, 6.2228025878281625469d-11,  &
     &    -8.8636681342142794023d-10, -2.3661241616744818608d-8,  &
     &    -1.6777870960740520557d-6, -3.2189402882677074318d-4,   &
     &    5.1503226479551959376d-2 /                               
      data (c(i), i = 36, 44) /                                   &
     &    -4.5801527369223291722d-14, 6.7998819697143727209d-13,  &
     &    -4.1624857909290468421d-12, -3.2849009406112440998d-11, &
     &    -3.2478275690431118270d-10, -2.5739209934053714983d-8,  &
     &    -1.6730566573215739195d-6, -3.2190010909008684076d-4,   &
     &    5.1503229866932077150d-2 / 
      w = abs(x)
      if (w .lt. 8.5d0) then
          t = w * w * 0.0625d0
          k = 12 * (int(t))
          y = (((((((((((a(k) * t + a(k + 1)) * t +               &
     &        a(k + 2)) * t + a(k + 3)) * t + a(k + 4)) * t +     &
     &        a(k + 5)) * t + a(k + 6)) * t + a(k + 7)) * t +     &
     &        a(k + 8)) * t + a(k + 9)) * t + a(k + 10)) * t +    &
     &        a(k + 11)) * w
      else if (w .lt. 12.5d0) then
          k = int(w)
          t = w - k
          k = 14 * (k - 8)
          y = ((((((((((((b(k) * t + b(k + 1)) * t +              &
     &        b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t +     &
     &        b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t +     &
     &        b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t +    &
     &        b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
      else
          t = 60 / w
          k = 9 * (int(t))
          y = ((((((((c(k) * t + c(k + 1)) * t +                  &
     &        c(k + 2)) * t + c(k + 3)) * t + c(k + 4)) * t +     &
     &        c(k + 5)) * t + c(k + 6)) * t + c(k + 7)) * t +     &
     &        c(k + 8)) * sqrt(t) * exp(w)
      end if
      if (x .lt. 0) y = -y
      dbesi1 = y
      end function dbesi1
!


END MODULE GKV_math
